#! /usr/bin/perl

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States. Recipients
# of this software assume all responsibilities associated with its operation,
# modification and maintenance. 

use strict;
use TrialsCBCD09;
use MetricTV08;
use Data::Dumper;
use DETCurve;
use DETCurveSet;
use BipartiteMatch;

#############################################################################
# evaluateCopyDetectionSubmission.pl
#
# The purpose of this program is to evaluate TRECVID 2008 copy detection
# submissions according to the rules spelled out in the guidelines: 
# http://www-nlpir.nist.gov/projects/tv2008/Evaluation-cbcd-v1.3.htm#eval
#
# Inputs:
#
#   - a submission file containing one run
#   - original ground truth file
#   - audio ground truth file
#   - video ground truth file
#   - audiovideo ground truth file
#
# Outputs:
#
#   - a results detail file with info about how each submission line was processed
#   - a results table with summary counts per transformation of total queries, total 
#     processing time, misses, false alarms, true positives, the minimal normalized 
#     detection cost rate (NDCR), the optimal decision score threshold associated with 
#     the minimal NDCR, mean processing time of true positive queries, mean F1-score 
#     of true positive queries.	  
#   - a DET curve (.png) per transformation: probability of miss versus false alarm 
#     rate (results for all queries using that transf.)
#     See: http://www.nist.gov/speech/publications/storage_paper/det.pdf 
#
# Usage:
#   perl -I common/lib evaluteSingleCBCD.pl SUBMISSIONFILELIST
#   
#############################################################################

# Evaluate a validated copy detection submission

my($slash) = "/";
my($originalGroundTruth) = "./ground.truth/o.gt";
my($audioGroundTruth) = "./ground.truth/a.gt";
my($videoGroundTruth) = "./ground.truth/v.gt";
my($audiovideoGroundTruth) = "./ground.truth/av.gt";

# Header text: Total Queries, Mean F1, Mean processing time, Total processing time, True positive count, Miss count, False alarm count, Min normalized DCR, optimum detection score
my(@headers) = ("Total_Queries" , "Mean_proc_time", "Total_proc_time", "Opt_Mean_F1" , "Opt_TP_count", "Opt_Miss_count", "Opt_SubMiss_count","Opt_FA_count", "Opt_TN_count", "Opt_Min_NDCR", "Decision_Threshold", "Opt_Min_Rfa", "Opt_Min_Pmiss", "Act_NDCR","Act_Rfa","Act_Pmiss","Act_TP_count", "Act_Miss_count","Act_SubMiss_count", "Act_FA_count","Act_TN_count","Act_Mean_F1", "Type3_FAcount"); 

# Detail file headers: Query ID, Transformation no., True copy(1|0), Judgment(1|0), detection score, recall, precision, F1-score
my(@headerString) = ("\#Query ID","   |   #Transformation number","   |   | #Truecopy (1 if query copntains a copy (type 1 or 2), 0 otherwise)","   |   | | #Judgment (1 if correct detection, 0 otherwise)","   |   | | |  #decision score","   |   | | |  |     #recall ","   |   | | |  |     |      #precision ","   |   | | |  |     |      |      #F1 localization score for TPs","   |   | | |  |     |      |      |          #submitted Query Id","   |   | | |  |     |      |      |          |        #submitted ref. video Id", "   |   | | |  |     |      |      |          |        |	               #first ref. video time-code", "   |   | | |  |     |      |      |          |        |                |	   #last ref. video time-code","   |   | | |  |     |      |      |          |        |                |           |		#decision score", "   |   | | |  |     |      |      |          |        |                |           |            |	    #first query time-code", "   |   | | |  |     |      |      |          |        |                |           |            |           |");

############################################################

# bottom and top Query IDs for the three run types
my($botValidQueryId_a) = 3001;
my($topValidQueryId_a) = 4407;

my($botValidQueryId_v) = 1;
my($topValidQueryId_v) = 1407;

my($botValidQueryId_m) = 5001;
my($topValidQueryId_m) = 14849;

my($botValidQueryId);
my($topValidQueryId);

my($audio_transformations_num) = 7;
my($video_transformations_num) = 10;

#[Tquery per transformation] * [Trefdata]
my($durationsTotal) = 0;

#total Ref. durations in hr.
my ($Refduration) = 380;

#Initialized parameters for DET calculations, Rtarget, Miss cost, FA cost
my($profile);
my($rtarget) = 0.005;
my($cmiss) = 10;
my($cfa) = 1;
my($run_threshold);
my($minDS) = 1000.0;
my($maxDS) = 0.0;
my($rangeDS);

############# two hashes used by alignment method
my %sysobj = ();
my %refobj = ();
my @kp = ();
my $sys_mobj;
my $ref_mobj;
my $jv;
my $bpm_judg;
##################################################
my($line);
my($code);
my($rest);
my($firstav);
my($oav);
my($type);
my($junkFile) ;
my($junkStart);
my($junkEnd) ;
my($refVid) ;
my($refStart);
my($refStop) ;
my($queryDuration);
my($offset) ;
my($h) ;
my($m) ;
my($s) ;
my($n) ;
my($d) ;
my($orig) ;
my($transformed) ;
my($transformation) ;
my($runFile) ;
my($rc) ;
my($runType) ;
my($i) ;
my($j);
my($T) ;
my($maxT) ;
my($qtype) ;
my($t) ;
my($indexT) ;
my($countT);
my($groupCount);
my($qId) ;
my($vId) ;
my($firstRefTime) ;
my($lastRefTime) ;
my($conf) ;
my($firstQueryTime) ;
my($judgment);
my($maxfscore);
my($recall);
my($precision);
my($fscore);
my($tmpjudgment);
my($copy);
my($processTime);
my($detCurveName);
my($findex);
my($maxfindex);
my($allMiss);
my($transformedAudio);
my($transformedVideo);
my($transformedAudioVideo);
my($v_t);
my($a_t);


my($tmpqId);
my($tmpvId);	
my($tmpfirstRefTime);
my($tmplastRefTime);
my($tmpconf);
my($tmpfirstQueryTime);
###################################################

my(@line);
my(@type);
my(@refVideo);
my(@refStart);
my(@refStop);
my(@queryDuration);
my(@trans2orig);
my(@transformation);
my(@unsortedRun);
my(@sortedRun);
my(@trialArr);
my(@trial_FA_Arr);
my(@rItems);
my(@resultsArr);
my(@judgmentSet);
my(@tmplist);

my(@proccessTimeArr);
my(@countFA);
my(@meanF1);
my(@countTP);
my(@meanPTime);
my(@countMiss);
my(@act_meanF1);
my(@act_TP);
my(@act_FA);
my(@act_MISS);
my(@act_MISS_submitted);
my(@opt_TP);
my(@opt_FA);
my(@opt_MISS);
my(@opt_MISS_submitted);
my(@opt_meanF1);
my(@act_TN);
my(@opt_TN);

my(@detScores);
my(@processTimeTotal);
my(@totalTqueries);
my(@allMissArr);
my(@cont2);
my(@allF1);
my(@av_trans_matrix);
my(@trans2avOrig);

my @isolinecoef = ( );
my($File);

$runFile = $ARGV[0];

$i = rindex($runFile,$slash);
if ($i != -1)
{
    $File = substr($runFile,$i+1);
}
else
{
    $File = $runFile;
}

#my $File = substr($runFile,rindex($runFile,$slash)+1);

open RESULTS, ">./results/analysis/$File.results.table" ||
    die "Can't open results.table: $!\n";

open DIAGNOSTIC, ">./results/analysis/$File.results.detail" ||
    die "Can't open results.detail: $!\n";


#create new set of DET curves object
my $ds = new DETCurveSet("DETS for Submission $File");
####################################################################################
######################## Initializations for retrieving GT data ####################

# Read in the info on untransformed queries,their types, sources
# Basically the objective is to be able to retrieve the transformation, type, and total duration of each query
# in addition to the starting and ending points of the reference video (used to calculate the F1-score)

open OGT, "$originalGroundTruth" or
    die "Can't open in $originalGroundTruth: $!\n";

    while ($line = <OGT>)
    {

	(undef,$firstav,$oav,$type,$junkFile,$junkStart,$junkEnd,$refVid,$refStart,$refStop,$offset) = split /\s+/," $line";
	$refVid = substr($refVid,rindex($refVid,$slash)+1);
	$type[$oav] = $type;
	$refVideo[$oav] = $refVid;
	($h,$m,$s,$n,$d,undef) = split /[:\/]/,$refStart;
	$refStart = (3600*$h)+(60*$m)+$s;
	if ($d != 0)
	{
	    $refStart += $n/$d;
	}
	$refStart[$oav] = $refStart;
	#print "START $refStart - $h $m $s $n $d = $refStart\n";

	($h,$m,$s,$n,$d,undef) = split /[:\/]/,$refStop;
	$refStop = (3600*$h)+(60*$m)+$s;
	if ($d != 0)
	{
	    $refStop += $n/$d;
	}
	$refStop[$oav] = $refStop;
	#print "STOP $refStop - $h $m $s $n $d = $refStop\n";

	#  Get whole duration of each query #############################
	# this will be used by the trial object

	($h,$m,$s,$n,$d,undef) = split /[:\/]/,$junkStart;
	$junkStart = (3600*$h)+(60*$m)+$s;
	if ($d != 0)
	{
	    $junkStart += $n/$d;
	}

	($h,$m,$s,$n,$d,undef) = split /[:\/]/,$junkEnd;
	$junkEnd = (3600*$h)+(60*$m)+$s;
	if ($d != 0)
	{
	    $junkEnd += $n/$d;
	}
	$queryDuration[$oav] = ($refStop - $refStart) + ($junkEnd - $junkStart) + 2;
	
	########### Added in tv10 to calc. Tquery / transformation (total query durations)
	$durationsTotal += $queryDuration[$oav];
    }

# Duration per Transformation (in hours) * Duration of Ref. data
$durationsTotal = $durationsTotal / 3600;
$durationsTotal *= $Refduration; 

# Read info on transformed audio queries, original versions, transformation(s)
open AGT, "$audioGroundTruth" or
    die "Can't open $audioGroundTruth: $!\n";
while ($line = <AGT>)
{
    (undef,$orig,undef,$transformed,$transformation) = split /\s+/," $line";
    if ($transformed != 0)
    {		
    	$trans2orig[$transformed] = $orig;
        $transformation[$transformed] = $transformation;
    }	
}


# Read info on transformed video queries, original versions, transformation(s)
open VGT, "$videoGroundTruth" or
    die "Can't open $videoGroundTruth: $!\n";
while ($line = <VGT>)
{
    (undef,$orig,$transformed,$transformation) = split /\s+/," $line";
    $trans2orig[$transformed] = $orig;
    $transformation[$transformed] = $transformation;
}



########## comment for only testing tv2009 without av.gt , otherwise un-comment

# Read info on transformed video queries, original versions, transformation(s)
 open AVGT, "$audiovideoGroundTruth" or
    die "Can't open $audiovideoGroundTruth: $!\n";

#create a transformation ID matrix(1-70) indexed by the video and audio transformation IDs. (to remove gaps between seq. of numbers of combined #transformation Ids of video and audio)

# array to store the IDs of the dropped transformations.
my(@dropped);
my($count) = 1;
for ($i=1; $i <= $video_transformations_num; $i++)
{
	for ($j=1; $j <= $audio_transformations_num; $j++)
	{
		$av_trans_matrix[$i][$j] = $count;
		
		#to be able to retrieve the original audio and video transformations
	        $trans2avOrig[$count][1] = $i;
	        $trans2avOrig[$count][2] = $j;
		
		#TV2009 T1, T7 and T9 were dropped. Here we are keeping track of the corresponding m Transformation IDs.
		if ($i == 1 || $i == 7 || $i == 9)
		{
			$dropped[$count] = -1;			
		}	
		$count++;
	}
}

while ($line = <AVGT>)
{
    (undef,$orig,$transformedAudio,$transformedVideo,$transformedAudioVideo,$v_t,$a_t) = split /\s+/," $line";

   $trans2orig[$transformedAudioVideo] = $orig;
    	
   $transformation[$transformedAudioVideo] = $av_trans_matrix[$v_t][$a_t];

}

############################# END of initialization of GT data #################




################################################################################
# call the validateSubmission first to make sure it is a valid run file
#system('cd submission.checker');
#my($testrc) = system('tv9.check_copydetection.pl',$runFile);
#print "reurned val = $testrc \n";
#if (system('tv9.check_copydetection.pl',$runFile) != 0)
#{
#	print "Not a valid submission! \n";
#	$rc = -1;
#	exit $rc;
#}
#system('cd ..');

################################################################################
# open run file
open RUN, "$runFile" or
    die "Can't open runfile \"$runFile\": $!\n";

#sort results by Query ID (this is to make sure that multiple submitted result items are sequentially in the submitted file)
@unsortedRun = <RUN>;
@sortedRun = sort @unsortedRun;


###############################################################################
# Derive run type from file name: GROUPID.RUNTYPE.runId
# where RUNTYPE = a | v | m

$i = rindex($runFile,$slash);
if ($i != -1)
{
    $runFile = substr($runFile,$i);
}

################################################## we have to add ,undef to the split to accomodate the "profile" in the run file name for TV2009
(undef,$runType,undef) = split /\./,$runFile;

$runType = uc $runType;

# Assign bottom and top Query IDs based on the run type
if ($runType eq "A")
{
	$botValidQueryId = $botValidQueryId_a;
	$topValidQueryId = $topValidQueryId_a;
        $maxT = $audio_transformations_num + 1;

}
elsif ($runType eq "V")
{
	$botValidQueryId = $botValidQueryId_v;
	$topValidQueryId = $topValidQueryId_v;
	$maxT = $video_transformations_num + 1;
}
elsif ($runType eq "M")
{
	$botValidQueryId = $botValidQueryId_m;
	$topValidQueryId = $topValidQueryId_m;
        $maxT = $video_transformations_num * $audio_transformations_num + 1;
}


########################################################
#1
# For each transformation
    # Build an initial result set containing an empty entry for each
    # TP query from the current transformation. This is based solely
    # on the ground truth

#2
# Ritems is initialized to store the count of each group of multiple result items

#3 
# The total durations of all the queries are summed and used later by the DET curve functions
for ($i = $botValidQueryId; $i <= $topValidQueryId; $i++)
{
	$T = getQueryTransformation($i);
        $qtype = getQueryType($i);
	# if Query is a "COPY"
        if ($qtype != 3) 
	{
	   	$trialArr[$T][$i] = ();

	}
	else
	{
		#Array of type 3 queries (used to detect if there is submissions from this type)
		$trial_FA_Arr[$T][$i] = ();

	}

	# initialize counter of submitted results per query
	$rItems[$T][$i] = 0;
	$allMissArr[$i] = 0;
	
}
# END For each transformation
########################################################
#1
# store each result item in a result set array (resultsArr) indexed by the transformation for later use

#2
# store the count of result items for each query

#3
# store the processing time (proccessTimeArr) of each query indexed by transformation and QID


# For each item in the result set (read from the sorted file)

  # if item is a submitted result item
    # Look up the transformation associated with this query
    # Add current result item to the resultArr indexed by its transformation
    # Count the number of result items for this query

  # if item is a processing time
    # Look up the transformation associated with this query
    # Add current processing time to proccessTimeArr indexed by transformation & query Id
 # END For each item in the result set

#my($syskey) = 1;

my($qrefid);

foreach $line(@sortedRun)
{

    chomp $line;
    $line =~ s/\r//g;

    (undef,$code,$rest) = split /\s+/," $line",3;

    if ($code eq "R")
    {
	($qId,$vId,$firstRefTime,$lastRefTime,$conf,$firstQueryTime) = split /\s+/,$rest;
	$T = getQueryTransformation($qId);
	$qrefid = getRefVideo($qId);
	#print "query : $qId Transf : $T  Ref. video :  $qrefid\n";
	$rItems[$T][$qId]++;
	push @{ $resultsArr[$T] }, [$rest];
        $totalTqueries[$T]++;
	if ($conf < $minDS) {$minDS = $conf;}
	if ($conf > $maxDS) {$maxDS = $conf;}

	#if ($T == 1) {print("$line\n");}	
	
    }
    if ($code eq "T")
    {
	($qId, $processTime) = split /\s+/,$rest;
	$T = getQueryTransformation($qId);
	$proccessTimeArr[$T][$qId] = $processTime;
	$processTimeTotal[$T] = $processTimeTotal[$T] + $proccessTimeArr[$T][$qId];

	#if ($T == 1) {print("$line\n");}
    }		
    if ($code eq "P")
    {
 	($profile) = split /\s+/,$rest;
	if ($profile eq "NOFA")
	{
		$cmiss = 1;
		$cfa = 1000;
	}
	elsif ($profile eq "BALANCED")
	{
		$cmiss = 1;
		$cfa = 1;
	}
    }

    if ($code eq "V")
    {
	($run_threshold) = split /\s+/,$rest;
	#print ("th = $run_threshold \n");
    }	

    
}


######################################################## test calling ernel function

########################################################3




#range of decision scores for this run
$rangeDS = $maxDS - $minDS;

#print the header information for the detailed result file
for ($i = 0; $i < 15; $i++)
{
	print DIAGNOSTIC "$headerString[$i]\n";
}

#retrieves the number of transforms (by getting the length of the array of lists resultsArr)
#$maxT = scalar(@resultsArr);


# For each transformation in the result set (resultsArr)
for ($t = 1; $t < $maxT ; $t++)
{

	#counter to count number of type 3 queries that were submitted as true copies!
	$cont2[$t] = 0;

   if ($resultsArr[$t] != ())
   {

   	#get how many results are in this transform	list
   	$countT = scalar(@{$resultsArr[$t]});

   }
   else
   {
 	$countT = 0;
   }

   #initialize the index that we will use to traverse the list
   $indexT = 0;

   while ($indexT < $countT)
   {
	# reads current line of query result
	@line = (@{ $resultsArr[$t][$indexT] });
	
     	($qId,$vId,$firstRefTime,$lastRefTime,$conf,$firstQueryTime) = split /\s+/,"@line";
	
	# For each group of multiple submission items for same query
	if ($rItems[$t][$qId] > 1)
	{
	   # Initialize the maxfscore (store max. fscore in it)
	   # tmplist (stores judgment data for this group)
           $maxfscore = 0;
	   $maxfindex = -1;	
	   $allMiss = 0;	
	   @tmplist = ();
	   
	   %sysobj = ();
	   %refobj = ();
	   $jv = 0;		
	   #############################################################################
	   # store information about the ref observation to be used in the new alignment method TRECVID2009
   	  
		########## may be need to only fill Ref in case of queries of type 1 or 2 only
		
		if (getQueryType($qId) != 3)
		{ 
			$refobj{"0"} = [$qId, getRefVideo($qId), getRefStart($qId), getRefStop($qId)];
		}

	   #############################################################################	

	   #calc. max. index in this transformation for that group
	   $groupCount = $indexT + $rItems[$t][$qId] - 1;
	
	   # For each result item in a group
	   # findex is the index used to store the location of max. fscore found.
	   for ($findex = 0; $indexT <= $groupCount; $indexT++, $findex++)
	   {

           	# reads current line of query result
		@line = (@{ $resultsArr[$t][$indexT] });
		
     		($qId,$vId,$firstRefTime,$lastRefTime,$conf,$firstQueryTime) = split /\s+/,"@line";

		########################################################################
		# store information about system observation to be resolved later by BPM algorithm for new alignment method TRECVID2009
		if (getQueryType($qId) != 3)
		{
			$sysobj{$findex} = [$qId,$vId,$firstRefTime,$lastRefTime,$conf];
		}
		#########################################################################

                # Call getJudgment passing queryId, videoId, submitted extent
	        # to get a judgment , F1 score, recall, precision and if it is a true copy or not
                ($judgment, $fscore, $recall, $precision, $copy) = getJudgment($qId, $vId, $firstRefTime, $lastRefTime, $conf);
		
	        # If judgment is 0
                  # add item in temp. list
		if ($judgment == 0)
                {
		    push @tmplist, [0, $conf, $fscore, $recall, $precision, $copy, @line];
		    $allMiss++;
			
		}
                # If judgment is 1
                  # add item to temp. list
	  	  # remember the greatest F score so far
                else
                {
		    push @tmplist, [1, $conf, $fscore, $recall, $precision, $copy , @line];
		    if ($fscore > $maxfscore) 
		    {
			$maxfscore = $fscore;
			$maxfindex = $findex;
		    }
                }
	   } #for

	   ####################################################################################
	 if (getQueryType($qId) != 3)
	 {
	  	#print("begin to apply BPM...........\n");
		my $bpm = new BipartiteMatch(\%refobj, \%sysobj, \&kernel_function, \@kp);
        	MMisc::error_quit("While creating the Bipartite Matching object for query ($qId) (" . $bpm->get_errormsg() . ")") if ($bpm->error());

		$bpm->compute();
		die("While computing the Bipartite Matching (" . $bpm->get_errormsg() . ")") if ($bpm->error());

		#print("finished computing of query $qId in T = $t\n");

		#$bpm->_display("joint_values");
		#$bpm->_display("mapped", "unmapped_ref", "unmapped_sys");


		my @mapped = $bpm->get_mapped_ids();
	      	MMisc::error_quit("Problem obtaining the mapped objects from the BPM (" . $bpm->get_errormsg() . ")") if ($bpm->error());

		if (scalar @mapped == 0) 
		{
			print "NO MAPPED\n";
		} 
		else 
		{
		        foreach my $mop (@mapped) 
			{
			#store the "key" of the mapped ref & sys..later we will compare this key to the current query to find out if we should put 0 or 1 in judgment..we will put 1 when the current query maps to the key of the mapped sys here.
		        ($sys_mobj, $ref_mobj) = (@{$mop});
		#	print "SYSID: $sys_obj / REFID: $ref_obj\n";
		        my ($tsysQID, $tsysVideo, $tsysStart, $tsysStop, $tsysConf) =  @{$sysobj{$sys_mobj}};
		        my ($trefQID, $trefVideo, $trefStart, $trefStop, $trefConf) =  @{$refobj{$ref_mobj}};
			$jv = $bpm->get_jointvalues_refsys_value($ref_mobj, $sys_mobj);
			#if ($jv != 0) {print "SYS: $tsysVideo match REF: $trefVideo.....sys id key = $sys_mobj \n";}
			}
		}
	 }

	   ##############################################################################################	



       	   # For each item in list
       	     # if item has max F score
               # set judgment in judgment set to 1 , AND store the rest of the judgment data (conf., fscore, recall, precision)
             # else
               # set judgment in judgment set to 0 , AND store the rest of the judgment data (conf., fscore, recall, precision)
             # if query is true copy
	       # set judgment, conf., and fscore in trialArr for the current transform

           for ($tmpjudgment = 0; $tmpjudgment < $rItems[$t][$qId]; $tmpjudgment++)
           {
		############################################## Selecting the mapped sys to set its judgment to 1 in TRECVID2009 (new alignment method)
	
		if (($jv != 0) && ($tmpjudgment == $sys_mobj))
		{
			$bpm_judg = 1;
			#print("from f-score loop ---> $tmpjudgment \n");
		}
		else
		{
			$bpm_judg = 0;
		}
		##############################################

	       	
	       #if judgment == 0, set result set = 0 and add the rest of the item information
	       if ($tmplist[$tmpjudgment][0] == 0)
	       {
					   #query Id,   #judg,     #detection score ,  #fscore , #recall  , #precision, #truecopy, #submision result item
		   push @{ $judgmentSet[$t] }, [$qId, $bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2], $tmplist[$tmpjudgment][3], 
						$tmplist[$tmpjudgment][4], $tmplist[$tmpjudgment][5], $tmplist[$tmpjudgment][6]];

		   #if it is a true copy query
		   if (getQueryType($qId) != 3)		
		   {   
			     			
							#judgment  , #detection score, #fscore
		       push @{ $trialArr[$t][$qId] } , [$bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2]];
					
		   }
		   else 
		   {
		       #count FA of type 3
		       $cont2[$t]++;
		       #add a Flag for this query
				
		       push @{ $trial_FA_Arr[$t][$qId] } , [$tmplist[$tmpjudgment][1]];
		   }

		   $countFA[$t]++;
						
	       }           
	       else
	       {
		   #if judgment == 1, look for the max fscore
		   if ($tmplist[$tmpjudgment][2] == $maxfscore)
		   {	
  						
		       #check if it is the first element with max. F-score
		       if ($maxfindex == $tmpjudgment)
		       {
						      #query Id,   #judg,     #conf ,  #fscore , #recall  , #precision,  #truecopy, #submision result item
			   push @{ $judgmentSet[$t] }, [$qId, $bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2], $tmplist[$tmpjudgment][3], 
							$tmplist[$tmpjudgment][4], $tmplist[$tmpjudgment][5], $tmplist[$tmpjudgment][6]];

			   #if it is a true copy query
			   if (getQueryType($qId) != 3)		
			   {
									#judgment  , #detection score , #fscore
			       push @{ $trialArr[$t][$qId] } , [$bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2]];
					
			       #Add the query fscore and count a true positive
			       $meanF1[$t] += $tmplist[$tmpjudgment][2];
			       $countTP[$t]++;
														
			   }
		       }				
			
		       else
		       {
			   push @{ $judgmentSet[$t] }, [$qId, $bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2], $tmplist[$tmpjudgment][3], 
							$tmplist[$tmpjudgment][4], $tmplist[$tmpjudgment][5], $tmplist[$tmpjudgment][6]];

			   #if it is a true copy query
			   if (getQueryType($qId) != 3)		
			   {
			       #judgment  , #detection score, #fscore
			       push @{ $trialArr[$t][$qId] } , [$bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2]];	
					
			   }

			   #count a false alarm
			   $countFA[$t]++;
		       }
		   }	
		   else #if fscore != maxfscore
		   { 
 						      #query Id,   #judg,     #conf ,  #fscore , #recall  , #precision, #copy, #submision result item
		       push @{ $judgmentSet[$t] }, [$qId, $bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2], $tmplist[$tmpjudgment][3], 
						    $tmplist[$tmpjudgment][4], $tmplist[$tmpjudgment][5], $tmplist[$tmpjudgment][6]];

		       #if it is a true copy query
		       if (getQueryType($qId) != 3)		
		       {
			   #judgment  , #detection score, #fscore
			   push @{ $trialArr[$t][$qId] } , [$bpm_judg, $tmplist[$tmpjudgment][1], $tmplist[$tmpjudgment][2]];	
					
		       }

		       #count a false alarm
		       $countFA[$t]++;
		   }
	       } #else
		($tmpqId,$tmpvId,$tmpfirstRefTime,$tmplastRefTime,$tmpconf,$tmpfirstQueryTime) = split /\s+/,"$judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][7]";

	       #print on each line
	       #Qid #T #Truecopy #Judgment #conf #recall #precision #fscore #submission result item
	       printf DIAGNOSTIC "%5s %2d %1s %1s %6.3f %6.4f %6.4f %6.4f   %6s   %14s   %9s   %9s   %8s   %9s \n" ,  $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][0], $t, $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][6], $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][1], 
	       $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][2], $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][4], 
	       $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][5] , $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][3], 
		$tmpqId,$tmpvId,$tmpfirstRefTime,$tmplastRefTime,$tmpconf,$tmpfirstQueryTime;
	
	   } #for

	   #Check if all submitted result items were FA..in this case we have to set a Miss in the trial object
	   if (($rItems[$t][$qId] == $allMiss) && (getQueryType($qId) != 3))
	   {
	       #set a flag indexed by the qId 
	       $allMissArr[$qId] = 1;
	        
	   }


       } # END For each group of multiple result items for same query

	else
	{
	    # For each unjudged item in the result set
	    # call getJudgment to get true judgment
	    # set judgment, conf., fscore, recall, and precision in judgment set to true judgment
	    # if Query is true copy
	       # set judgment, conf., and fscore in trialArr for the current transform

            ($judgment, $fscore, $recall, $precision, $copy) = getJudgment($qId, $vId, $firstRefTime, $lastRefTime, $conf);
            push @{ $judgmentSet[$t] }, [$qId, $judgment, $conf, $fscore, $recall, $precision, $copy];

	    #if it is a true copy query	
	    if (getQueryType($qId) != 3)	
	    {	
						#judgment  , #detection score, #fscore
		push @{ $trialArr[$t][$qId] } , [$judgment, $conf, $fscore];

		if ($judgment == 1)
		{
		   #Add fscore and count a True Positive	
		   $meanF1[$t] += $fscore;
  		   $countTP[$t]++;
 		   		
		}
			
	    }
	    else { 
		 #count type 3 FA queries	
		 $cont2[$t]++; 
 		 push @{ $trial_FA_Arr[$t][$qId] } , [$conf];
		 }

	  
	    if ($judgment == 0)
	    {
		#count a false alarm
		$countFA[$t]++;

		if (getQueryType($qId) != 3)
	        {
		    #set a flag indexed by the qId to indicate a Miss
		    $allMissArr[$qId] = 1;
	            
	        }
	    }	
	    
    	    ($tmpqId,$tmpvId,$tmpfirstRefTime,$tmplastRefTime,$tmpconf,$tmpfirstQueryTime) = split /\s+/,"@line";
	    #print on each line
	    #Qid #T #Truecopy #Judgment #conf #recall #precision #fscore #submission result item
	    printf DIAGNOSTIC "%5s %2d %1s %1s %6.3f %6.4f %6.4f %6.4f   %6s   %14s   %9s   %9s   %8s   %9s \n" ,  $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][0], 
  	    $t, $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][6], $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][1], 
	    $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][2], $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][4], 
	    $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][5] , $judgmentSet[$t][scalar(@{$judgmentSet[$t]})-1][3],
	    $tmpqId,$tmpvId,$tmpfirstRefTime,$tmplastRefTime,$tmpconf,$tmpfirstQueryTime;
			
            $indexT++;
	} #else  # END For each single unjudged item in the result set

	
   } #END of current transformation



################################################################################
########################## Adding trials

my($tmpx);
my($j);

   # get a new Trials object passing:
   #print "Total query duration: $durationsTotal\n"; exit;
   my $tString = sprintf("%03d",$t);
   
   # total query duration in seconds (will be converted to hours in DET-curve code)
   my $trialT = new TrialsCBCD09("Copy Detection", "Transformation $t", "Occurrence", { ("TOTALDURATION" => $durationsTotal) });

# For each item in the trial (trialArr)
	 # if (item is a miss, i.e. has empty (judgment,ds) list)
	    #   call addTrial to current Trials object, passing:
	    #     blockID = runId+Transformation#
	    #     decision score = undef
	    #     optimal threshold = "NO"
	    #     true judgment = 1
         # else
	    #   call addTrial to current Trials object, passing:
	    #     blockID = runId+Transformation#
	    #     decision score = from result set
	    #     optimal threshold = "NO"
	    #     true judgment = from result set

##################################################################################
# Tabel of how we count the submitted results
#  Count 	Above (YES) / Below (NO) threshold		Target (Copy) /NonTarget (Non-copy)
#  TP		YES						Target
#  FA		YES						NonTarget
#  MISS1	NO						Target
#  MISS2	OMMIT (not applicable)				Target
#  TN		NO						NonTarget

#Then the total counts for Ref = TP + MISS1 + MISS2
#     the total counts for Sys = TP + FA + TN + MISS1

# this can be done on the optimum threshold and the actual threshold to get a set of counts for each case.	
##################################################################################



my($cont) = 0;


   for ($i = $botValidQueryId; $i <= $topValidQueryId; $i++)
   {
	
  	$qtype = getQueryType($i);
	
        if (($qtype != 3) && (getQueryTransformation($i) == $t))
	{
	   #if it is not a miss
	   if ($trialArr[$t][$i] != ())
	   {
	     $tmpx = scalar(@{$trialArr[$t][$i]});
	   
	     for ($j = 0; $j < $tmpx; $j++)
	     {
		if ($run_threshold <= $trialArr[$t][$i][$j][1])
		{
					      		#detection score                  #judgment
			$trialT->addTrial("$File $t", $trialArr[$t][$i][$j][1], "YES", $trialArr[$t][$i][$j][0]);

			#count actual TP. "YES" and Target
			if ($trialArr[$t][$i][$j][0] == 1) 
			{
				$act_TP[$t]++; 
				$act_meanF1[$t]+= $trialArr[$t][$i][$j][2];
			}
			#count actual FA. "YES" and NonTarget
			elsif ($trialArr[$t][$i][$j][0] == 0) 
			{
				$act_FA[$t]++;
			}
		}
		else
		{
					      		#detection score                  #judgment
			$trialT->addTrial("$File $t", $trialArr[$t][$i][$j][1], "NO", $trialArr[$t][$i][$j][0]);	

			#count actual miss. "NO" and Target
			if ($trialArr[$t][$i][$j][0] == 1) 
			{
				$act_MISS_submitted[$t]++;
			}
			#count actual true negative. "NO" and NonTarget
			elsif ($trialArr[$t][$i][$j][0] == 0) 
			{
				$act_TN[$t]++;
			}
		}
		
		if ($trialArr[$t][$i][$j][0] == 0) {$cont++;}			
	        
	     }
	     # if all submitted results are FA, add 1 miss
	     if ($allMissArr[$i] == 1)	
	     {
		#a miss				  #true copy that didn't match any submitted result item
		$trialT->addTrial("$File $t", undef, "OMITTED", 1);
	
		#count actual miss. 
		$act_MISS[$t]++;		        

		#count a miss
		$countMiss[$t]++;
	     }

	   }
	   else
	   {	#a miss			     #true copy that was omitted  #TP query (thus 1) 	

		$trialT->addTrial("$File $t", undef, "OMITTED", 1);		

		#count a miss
		$countMiss[$t]++;		

		#count actual miss
		$act_MISS[$t]++;
	   }		  	
	}

	else 
	{ # If a result item belongs to type 3 we consider it as a FA
	  if (($qtype == 3) && (getQueryTransformation($i) == $t))
	  {
		if ($trial_FA_Arr[$t][$i] != ())
		{
			$tmpx = scalar(@{$trial_FA_Arr[$t][$i]});
	    		for ($j = 0; $j < $tmpx; $j++)
	     		{
				if ($run_threshold <= $trial_FA_Arr[$t][$i][$j][0])
				{
							   #detection score	  #system output    # FA (thus 0)
				    $trialT->addTrial("$File $t", $trial_FA_Arr[$t][$i][$j][0], "YES", 0);	

			  	    #count actual FA. "YES" and NonTarget
				    $act_FA[$t]++;
				}
				else
				{
							   #detection score	  #system output    # FA (thus 0)
				    $trialT->addTrial("$File $t", $trial_FA_Arr[$t][$i][$j][0], "NO", 0);

				    #count true negative. "NO" and NonTarget	
				    $act_TN[$t]++;	
				}
			}
		}
	  }
	}
	
		
    }

	#count2 is the number of type 3 queries that are submitted as copy candidates
    #print "T : $t , FA1 = $cont , FA2 = $cont2 \n";
	
    $allF1[$t] = $meanF1[$t]; 

    if ($countTP[$t] != 0) #avoid dividing by zero
    {		
    	    # calc. the mean F1 across all TP queries for that transform
	   
	    $meanF1[$t] = $meanF1[$t] / $countTP[$t];

    }
    else
    {
 	$meanF1[$t] = 0;

    }

    #Get number of original queries (array $type begins from zero, thus need to subtract 1)
    my($Qnum) = scalar(@type) - 1;
    $meanPTime[$t] = $processTimeTotal[$t] / $Qnum;
	    

    # call newDETCurve passing ....
    
    #print $trialT->dumpCountSummary();
    #my $transString = sprintf("%3d",$t);
    my ($detT);

    if ($runType eq "M")
    {
	
    	$detT = new DETCurve($trialT, 
    				    new MetricTV08({ ('CostMiss' => $cmiss, 'CostFA' => $cfa, 'Rtarget' => $rtarget ) }, $trialT),
    				    "Transformation $trans2avOrig[$t][1],$trans2avOrig[$t][2]", \@isolinecoef);
    }
    else
    {
	$detT = new DETCurve($trialT, 
    				    new MetricTV08({ ('CostMiss' => $cmiss, 'CostFA' => $cfa, 'Rtarget' => $rtarget ) }, $trialT),
    				    "Transformation $t", \@isolinecoef);
    }
	#print $trialT->dumpCountSummary();

	######################################################  NEW CODE FOR 2009 MODIFIED DET CODE
	print $detT->getMessages() if (! MMisc::is_blank($detT->getMessages()));

	######################################################


	#Store the Optimum detection score and Min DCR for each transform
   $detScores[$t][1] = $detT->getBestCombDetectionScore(); #Decision score
   $detScores[$t][2] = $detT->getBestCombComb(); #Min NDCR
   $detScores[$t][3] = $detT->getBestCombMFA(); #Rfa
   $detScores[$t][4] = $detT->getBestCombMMiss(); #Pmiss

	########################################################### new code for 2009 to get "actual" RFA, PMISS, and DCR
	my($MeanActComb);
	my($SampleStdDevActComb);
	my($MeanMiss);
	my($SampleStdDevMiss);
	my($MeanFA);
	my($SampleStdDevFA);

  ($MeanActComb, $SampleStdDevActComb, $MeanMiss, $SampleStdDevMiss, $MeanFA, $SampleStdDevFA) = $detT->{METRIC}->getActualDecisionPerformance();
	$detScores[$t][5] = $MeanActComb; #Actual DCR
	$detScores[$t][6] = $MeanFA; #Actual Rfa
	$detScores[$t][7] = $MeanMiss; #Actual Pmiss
	#print "Actual Rfa = $MeanFA , Actual Pmiss = $MeanMiss, Actual DCR = $MeanActComb\n";
	##########################################################
	

    my $ret1 = $ds->addDET("$tString", $detT); 
    if ($ret1 ne "success") {die "Error: $t $ret1"}


#######################################################################################################
################### Counting TP, F1, MISS and FA using the optimum threshold

   for ($i = $botValidQueryId; $i <= $topValidQueryId; $i++)
   {	
  	$qtype = getQueryType($i);
	
        if (($qtype != 3) && (getQueryTransformation($i) == $t))
	{
	   #if it is not a miss
	   if ($trialArr[$t][$i] != ())
	   {
	     $tmpx = scalar(@{$trialArr[$t][$i]});
	   
	     for ($j = 0; $j < $tmpx; $j++)
	     {
		#if conf. is > than the optimum threshold
		if ($detScores[$t][1] <= $trialArr[$t][$i][$j][1])
		{
			#count optimum TP. "YES" and Target
			if ($trialArr[$t][$i][$j][0] == 1) 
			{
				$opt_TP[$t]++; 
				$opt_meanF1[$t]+= $trialArr[$t][$i][$j][2];
			}
			#count optimum FA. "YES" and NonTarget
			elsif ($trialArr[$t][$i][$j][0] == 0) {$opt_FA[$t]++;}	
		}
		else
		{
			#count optimum miss. "NO" and Target
			if ($trialArr[$t][$i][$j][0] == 1) {$opt_MISS_submitted[$t]++;}	
				#count true negative. "NO" and NonTarget
			elsif ($trialArr[$t][$i][$j][0] == 0) {$opt_TN[$t]++;}
		}	
	     }
	     # if all submitted results are FA, add 1 miss
	     if ($allMissArr[$i] == 1)	
	     {	
		#count optimum miss
		$opt_MISS[$t]++;		        
	     }

	   }
	   else
	   {	
		#count optimum miss
		$opt_MISS[$t]++;
	   }		  	
	}

	else 
	{ # If a result item belongs to type 3 we consider it as a FA
	  if (($qtype == 3) && (getQueryTransformation($i) == $t))
	  {
		if ($trial_FA_Arr[$t][$i] != ())
		{
			$tmpx = scalar(@{$trial_FA_Arr[$t][$i]});
	    		for ($j = 0; $j < $tmpx; $j++)
	     		{
				if ($detScores[$t][1] <= $trial_FA_Arr[$t][$i][$j][0])
				{
			  	    #count optimum FA. "YES" and NonTarget
				    $opt_FA[$t]++;
				}
				else
				{
				    #count true negative. "NO" and NonTarget
				    $opt_TN[$t]++;
				}
			}
		}
	  }
	}			
    }

    if ($opt_TP[$t] != 0) #avoid dividing by zero
    {		
    	    # calc. the mean F1 across all TP queries for that transform
	   
	    $opt_meanF1[$t] = $opt_meanF1[$t] / $opt_TP[$t];

    }
    else
    {
 	$opt_meanF1[$t] = 0;

    }
    if ($act_TP[$t] != 0) #avoid dividing by zero
    {		
    	    # calc. the mean F1 across all TP queries for that transform
	   
	    $act_meanF1[$t] = $act_meanF1[$t] / $act_TP[$t];

    }
    else
    {
 	$act_meanF1[$t] = 0;

    }
#######################################################################################################

    print "Finished processing Transform $t\n";
   
}  # END For each transformation in result set

#########################################################################
# O U T P U T      F I L E S

# generates results.table file. For each run across all transforms:
# Mean F1 across all TP queries
# Mean processing time across all TP queries
# Min normalized DCR
# Threshold for Min normalized DCR

print "writing table file \n";

print RESULTS "Det Curve Parameters: \n";
print RESULTS "Rtarget : $rtarget \n";
print RESULTS "Cmiss : $cmiss \n";
print RESULTS "Cfa : $cfa \n\n";
print RESULTS "Run Threshold : $run_threshold \n\n$trialArr[$t][$i][$j][0]";


printf RESULTS "%55s","Transformations";
for ($t = 1; $t < $maxT ; $t++)
{
	if ((($t != 1 && $t != 7 && $t != 9) && ($runType eq "V")) || (($dropped[$t] != -1) && ($runType eq "M")) || $runType eq "A")
	{
	     printf RESULTS "        %8d", $t;
	}
}

print RESULTS "\n";

my($sumtp) = 0;
my($sumfa) = 0;
my($summiss) = 0;
#loop on the statistic items we want to print
for ($i = 0; $i < 23; $i++)
{	
    
   printf RESULTS "%30s", $File;
   printf RESULTS "%25s", $headers[$i];

   #loop on each transform
TLOOP:   for ($t = 1; $t < $maxT ; $t++)
   {
   	if (((($t == 1) || ($t == 7) || ($t == 9)) && ($runType eq "V")) || ($dropped[$t] == -1 && $runType eq "M")) 
	{
	  next TLOOP;
	}
	 
	     if ($i == 0) {printf RESULTS "        %8d", $totalTqueries[$t];} #Total no. of queries    
	     if ($i == 1) {printf RESULTS "        %8.2f", $meanPTime[$t];} #Mean processing time
	     if ($i == 2) {printf RESULTS "        %8d" , $processTimeTotal[$t];} #Total processing time
	     if ($i == 3) {printf RESULTS "        %8.3f", $opt_meanF1[$t];} #optimum Mean F1
	     if ($i == 4) {printf RESULTS "        %8d" , $opt_TP[$t];} #Opt. TP counts	
	     if ($i == 5) {printf RESULTS "        %8d" , $opt_MISS[$t];} #Opt. Misses counts	
	     if ($i == 6) {printf RESULTS "        %8d" , $opt_MISS_submitted[$t];} #
	     if ($i == 7) {printf RESULTS "        %8d" , $opt_FA[$t];} #Opt. FA counts
	     if ($i == 8) {printf RESULTS "        %8d" , $opt_TN[$t];} #nOpt. TN	
	     if ($i == 9) {printf RESULTS "        %8.3f", $detScores[$t][2];} #Min DCR	
	     if ($i == 10) {printf RESULTS "        %8.3f", $detScores[$t][1];} #Optimum detection score
	     if ($i == 11) {printf RESULTS "        %8.3f", $detScores[$t][3];} #optimum Rfa
	     if ($i == 12) {printf RESULTS "        %8.3f", $detScores[$t][4];} #optimum Pmiss
	     if ($i == 13) {printf RESULTS "        %8.3f", $detScores[$t][5];} #Actual DCR
	     if ($i == 14) {printf RESULTS "        %8.3f", $detScores[$t][6];} #Actual Rfa
	     if ($i == 15) {printf RESULTS "        %8.3f", $detScores[$t][7];} #Actual Pmiss
	     if ($i == 16) {printf RESULTS "        %8d" , $act_TP[$t];} #Actual TP counts	
	     if ($i == 17) {printf RESULTS "        %8d" , $act_MISS[$t];} #Actual Misses counts
	     if ($i == 18) {printf RESULTS "        %8d" , $act_MISS_submitted[$t];} #
	     if ($i == 19) {printf RESULTS "        %8d" , $act_FA[$t];} #Actual FA counts	
	     if ($i == 20) {printf RESULTS "        %8d" , $act_TN[$t];} #Act. TN
	     if ($i == 21) {printf RESULTS "        %8.3f", $act_meanF1[$t];} #actual Mean F1
	     if ($i == 22) {printf RESULTS "        %8d" , $cont2[$t];} #no. of type 3 queries that are submitted as True copies
	
   }
   print RESULTS "\n";
}

my($tempT) = $t;
print "Finished writing table file \n";


#close RESULTS;
#close DIAGNOSTIC;
#close RUN;


#printf RESULTS "\nTotal TP = %4d \n", $sumtp;
#printf RESULTS "\nTtotal Miss = %4d \n", $summiss;
#printf RESULTS "\nTotal FA = %6d \n", $sumfa;

#Generates DET curves for each run/transform
# to get PNG file of DET curve, minimum cost, threshold for minimum cost
# Save threshold, cost, and curve

##############Remove comment to draw DET curves
#print $ds->renderAsTxt("/home/barney1/trecvid/9/copy.detection/results/DET.files/$File.det", 1, 1, 
 #                         { (xScale => "log", Ymin => "0.00001", Ymax => "99", Xmin => "0.00001", Xmax => "100", 
 #                            BuildPNG => 1, ReportActual => 0, createDETfiles => 1 , DrawIsometriclines => 1,
#			      Isometriclines => [ MetricTV08::isoCombCoeffForDETCurve() ]) });

###################### New for 2009 for modified DET code
MMisc::writeTo("./results/DET.files/$File.det", ".scores.txt", 1, 0, 
               $ds->renderAsTxt("./results/DET.files/$File.det", 1, 1, 
                                    { (xScale => "log", Ymin => "0.00001", Ymax => "99", Xmin => "0.00001", Xmax => "100",
                                       BuildPNG => 1, createDETfiles => 1 , DrawIsometriclines => 1, 
				       Isometriclines => [ MetricTV08::isoCombCoeffForDETCurve() ]) }));

##########################################################################
# GetQtype(qID)
  # given a query ID, returns the type (1 | 2 | 3)
# END function

# Given the id of a transformed query, return its type number
sub getQueryType
{
    my ($transformedQueryId) = @_;
    return $type[$trans2orig[$transformedQueryId]];
}

###########################################################################
# GetTransformation(qID)
  # given a query ID, returns back the transformation number
# END function

# Given the id of a transformed query, return its transformation number (int. number)
sub getQueryTransformation
{
    my ($transformedQueryId) = @_;
    return $transformation[$transformedQueryId];
}
############################################################################

# Given the id of a transformed query, return start time of reference segment (seconds)
sub getRefStart
{
    my ($transformedQueryId) = @_;
    return $refStart[$trans2orig[$transformedQueryId]];
}
############################################################################

# Given the id of a transformed query, return stop time of reference segment (seconds)
sub getRefStop
{
    my ($transformedQueryId) = @_;
    return $refStop[$trans2orig[$transformedQueryId]];
}
#############################################################################

# Given the id of a transformed query, return the whole duration (seconds)
sub getQueryDuration
{
    my ($transformedQueryId) = @_;
    return $queryDuration[$trans2orig[$transformedQueryId]];
}


#############################################################################
# Given the id of a transformed query, return video source reference segment (video name)
sub getRefVideo
{
    my ($transformedQueryId) = @_;
    return $refVideo[$trans2orig[$transformedQueryId]];
}

###############################################################################
#### getJudgment evaluates the submitted result item. If the submitted video ID is the same
#### as the reference video ID then it calculates the recall, precision and F-score and set 
#### the judgment to 1. Otherwise it set judgment to 0 and return 0 values for recall, precision 
#### and F-score. $copy is a flag that is returned and indicates if that Query is actual copy query (type 1 or 2) or not.

# getJudgment(qID, vID, submitted_extent_start, submitted_extent_end)
  # calls GetGroundTruth(qID)
  # gets back query type, ref ID, ref extent start, ref extent end
  # initialize F1-score = 0
  # if (ref ID == vID)
    # calc. recall and precision
      # judgment = 1
      # calc. F1-score
  # else
    # judgment = 0
  # return judgment, F1-score
# END function

sub getJudgment
{
    #initializations
    my($judgment) = 0;
    my($fscore) = 0;
    my($refVideo);
    my($refStart);
    my($refStop);
    my($recall) = 0;
    my($precision) = 0;
    my($copy) = 0;
   	
    my($ET) = 1E-8;
    my($Eds) = 1E-6;	

    #submitted result data
    my ($qId, $vId, $extentStart, $extentStop, $confds) = @_;

    #if query judgment is not copy return 0 
    if (getQueryType($qId) == 3)
    {
	return ($judgment , $fscore, $recall, $precision, $copy); #(false alarm)
    }

    #get GT data
    $refVideo = getRefVideo($qId);
    $copy = 1;
    
    if ($refVideo ne $vId)
    {
	return ($judgment , $fscore, $recall, $precision, $copy); #miss
    }
    else
    {
 	$refStart = getRefStart($qId);
 	$refStop = getRefStop($qId);


 	$recall = &max((&min($extentStop,$refStop) - &max($extentStart,$refStart)),0) / ($refStop - $refStart);
 	$precision = &max((&min($extentStop,$refStop) - &max($extentStart,$refStart)),0) / ($extentStop - $extentStart);

	
        #we might need the fscore to get it back for users
        
	if ($precision + $recall == 0)
	{
		$fscore = 0;    #miss
		$judgment = 0;
		
	}
	else
	{
        	$fscore = (2 * $precision * $recall) / ($precision + $recall);  #match
		$judgment = 1;
		
	}
	return ($judgment , $fscore, $recall, $precision, $copy);
        
     }

}

##########################################################################
sub kernel_function
{
    #initializations
    my($fscore) = 0;
    my($refVideo);
    my($refStart);
    my($refStop);
    my($refQID);	
    my($sysVideo);
    my($sysStart);
    my($sysStop);
    my($sysQID);
    my($sysConf);
    my($recall) = 0;
    my($precision) = 0;
    my($kernel) = 0;		
    my($ET) = 1E-8;
    my($Eds) = 1E-6;	

   #print("Kernel_function...\n");

    #submitted result data
    my ($refobj, $sysobj, @params) = @_;

	if (! defined $refobj && defined $sysobj) {return ("",0)};	
	if (defined $refobj && ! defined $sysobj) {return ("",0)};

    ($refQID, $refVideo, $refStart, $refStop) =  split /\s+/,"@{$refobj}";

   #print("REF object in Kernel_function...$refQID, $refVideo, $refStart, $refStop \n");

    ($sysQID, $sysVideo, $sysStart, $sysStop, $sysConf) =  split /\s+/,"@{$sysobj}";

  #print("SYS object in Kernel_function...$sysQID, $sysVideo, $sysStart, $sysStop, $sysConf \n");


    
    #if query is not copy return 0 
    if (getQueryType($sysQID) == 3)
    {
	return ("",$kernel); #(false alarm)
    }

      
    if ($refVideo ne $sysVideo)
    {
	return ("",$kernel); #miss
    }
    else
    {
 	$recall = &max((&min($sysStop,$refStop) - &max($sysStart,$refStart)),0) / ($refStop - $refStart);
 	$precision = &max((&min($sysStop,$refStop) - &max($sysStart,$refStart)),0) / ($sysStop - $sysStart);

	
        #we might need the fscore to get it back for users
        
	if ($precision + $recall == 0)
	{
		#miss
		$kernel = 0;
	}
	else
	{
        	$fscore = (2 * $precision * $recall) / ($precision + $recall);  #match
		
		$kernel = 1 + $ET*$fscore + $Eds*(($rangeDS == 0) ? 1 : (($sysConf - $minDS) / $rangeDS));
	}
	return ("",$kernel);
        
     }


}


######################################################################################################
######## return min value 
sub min
{
    my ($a,$b) = @_;
    if ($a > $b)
    {
	return $b;
    }
    else
    {
	return $a;
    }
}

############# return max value
sub max
{
    my ($a,$b) = @_;
    if ($a > $b)
    {
	return $a;
    }
    else
    {
	return $b;
    }
}
